/*
 *  bose - a Lightweight Workflow Engine
 *  Copyright (C) 2006  Kunimasa Noda <kuni@pm9.com>
 *
 *  SPL - The SPL Programming Language
 *  Copyright (C) 2004, 2005  Clifford Wolf <clifford@clifford.at>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  bose_exec.c: Workflow Engine
 */

#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <mysql/mysql.h>


#ifndef USEWIN32API
#  include <sys/resource.h>
#endif

#include "spl.h"
#include "compat.h"
#include "spl_modules/mod_sql.h"

int got_sigint = 0;
int got_sigterm = 0;

// SNAPSHOTǼ (file / db)
char snapshot_mode;

// SNAPSHOTե̾
char snapshot_file[256];

// 顼ʸ
char *bose_error_dump;

// DB
struct	st_db_desc {
	char host[256];
	char port[256];
	char name[256];
	char user[256];
	char pass[256];
};

void db_desc_trim(char *buf) {
	char *p = buf;
	buf[255] = 0;
	while(*p > ' ') p++;
	*p = 0;
}

// SUSPEND
static struct spl_node *handler_flow_suspend(struct spl_task *t, void *d UNUSED)
{
	got_sigint = 1;
	return 0;
}

// SIGTERM䤤碌
static struct spl_node *handler_flow_get_sigterm_info(struct spl_task *t, void *d UNUSED)
{
	return SPL_NEW_INT(got_sigterm);
}

// ֥ץư
static struct spl_node *handler_flow_task_create(struct spl_task *t, void *d UNUSED)
{
	char *name = spl_clib_get_string(t);
	char *program = spl_clib_get_string(t);
	struct spl_node *ctx = spl_clib_get_node(t);

	struct spl_asm *as = spl_asm_create();
	as->vm = t->vm;

	if ( spl_compiler(as, program, "task_main", 0, 0) ) SPL_NEW_INT(1);
	if ( spl_compiler(as, "_flow_exit_subprocess(); task_kill();", "task_epilogue", 0, 0) ) SPL_NEW_INT(2);

	struct spl_task *task = spl_task_create(t->vm, name);

	if ( ctx ) {
		spl_put(task->vm, task->ctx);
		task->ctx = ctx;
	}

	spl_task_setcode(task, spl_asm_dump(as));
	spl_asm_destroy(as);

	task->flags |= SPL_TASK_FLAG_PAUSED;

	return SPL_NEW_INT(0);
}

// ֥ͭץΥ
static struct spl_node *handler_flow_count_subprocess(struct spl_task *t, void *d UNUSED)
{
	int count = 0;
	t = t->vm->task_list;
	while (t) {
		if ( t->id && !(t->flags & (SPL_TASK_FLAG_PAUSED | SPL_TASK_FLAG_ZOMBIE))) {
			count++;
		}
		t= t->next;
	}
	return SPL_NEW_INT(count);
}

// ֥ץ¸߳ǧ
static struct spl_node *handler_flow_isexist_subprocess(struct spl_task *t, void *d UNUSED)
{
	char *name = spl_clib_get_string(t);
	t = t->vm->task_list;
	while (t) {
		if ( t->id && strcmp(t->id, name)==0) {
			return SPL_NEW_INT(1);
		}
		t= t->next;
	}
	return SPL_NEW_INT(0);
}

// UNIXץIDμ
static struct spl_node *handler_flow_get_ppid(struct spl_task *t, void *d UNUSED)
{
	return SPL_NEW_INT(getppid());
}

// 顼ݡȳǼ
void bose_spl_report_func(int type, void *desc, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	bose_error_dump = spl_report_string(type, desc, fmt, ap);
	va_end(ap);
}

// SIGINTϥɥ
void sigint_handler(int dummy UNUSED)
{
	if (got_sigterm) return;
	got_sigint = 1;
	return;
}

// SIGTERMϥɥ
void sigterm_handler(int dummy UNUSED)
{
	got_sigterm = 1;
	return;
}

// DBǡ
char *my_escape_string(MYSQL *mysql, char *from) {
	char *to = (char*)malloc(strlen(from) * 2);
	mysql_real_escape_string(mysql, to, from, strlen(from));
	return to;
}

extern struct spl_node *handler_encode_mysql(struct spl_task *task, void *data UNUSED);
extern struct spl_node *sql_mysql_query_callback(struct spl_task *task, void *backend_data, const char *query);
extern void sql_mysql_open_callback(struct spl_task *task, struct spl_node *node, const char *data);
extern struct spl_node *handler_sql(struct spl_task *task, void *data UNUSED);
extern void handler_sqlnode(struct spl_task *task UNUSED, struct spl_vm *vm, struct spl_node *node, struct spl_hnode_args *args, void *data UNUSED);
extern struct spl_node *handler_encode_sql(struct spl_task *task, void *data UNUSED);
extern int sql_connect(struct spl_task *task, struct spl_node *node);

/**
 * This function creates an SQL database handle.
 *
 * The backend driver for the database must be loaded already when this
 * function is called. E.g.:
 *
 *	load "sql_sqlite";
 *	var db = sql_connect("sqlite", "");
 *
 * The database handler returned by this function may then be used as first
 * argument for the [[sql()]] function.
 *
 * An [[SqlEx]] exception is thrown on errors.
 */
// builtin sql_connect(driver_name, driver_data)
static struct spl_node *handler_sql_connect_modify(struct spl_task *task, void *data UNUSED)
{
	char *driver_name = spl_clib_get_string(task);
	char *driver_data = spl_clib_get_string(task);
	char *module, *driver;

	struct spl_node *conn = spl_get(0);

	my_asprintf(&module, "sql_%s", driver_name);
	if (strcmp(driver_name, "mysql")!=0) {
		spl_module_load(task->vm, module, 0);
		free(module);
	}

	conn->hnode_name = strdup("sql");

	my_asprintf(&driver, "%s:%s", driver_name, driver_data);
	spl_set_string(conn, driver);

	if (sql_connect(task, conn)) {
		spl_put(task->vm, conn);
		return 0;
	}

	return conn;
}


// եץSNAPSHOTǡΥꥹȥ
static char	sql[80960];
void restore_log(struct st_db_desc *db_desc, char *process_id) {
	MYSQL	*mysql, *con;
	MYSQL_RES	*res;
	MYSQL_ROW	row;
	int	err;
	char	uncompress1[16];
	char	uncompress2[16];

	mysql = mysql_init(NULL);
	if(NULL==mysql){
		fprintf(stderr, "error mysql_init\n");
		exit(1);
	}

	mysql_options(mysql, MYSQL_OPT_COMPRESS, 0);
	con = mysql_real_connect(mysql, db_desc->host,
		db_desc->user, db_desc->pass,
		db_desc->name, atoi(db_desc->port),
		NULL, 0);
	if (NULL==con){
		fprintf(stderr, "error mysql_real_connect\n");
		exit(1);
	}

	// comperssޥɤΥݡ̵ͭå
	if (strcmp(mysql_get_server_info(con), "4.1") >= 0) {
		strcpy(uncompress1, "uncompress(");
		strcpy(uncompress2, ")");
	} else {
		strcpy(uncompress1, "");
		strcpy(uncompress2, "");
	}

	// SNAPSHOT
	sprintf(sql, "select %sSNAPSHOT%s from tbl_BOSE_PROC_STAT where PROCESS_ID='%s'", uncompress1, uncompress2, process_id);
	err = mysql_query(mysql, sql);
	if (err != 0) {
		fprintf(stderr, "error mysql_query(%s)\n", sql);
		exit(1);
	}
	res = mysql_store_result(mysql);
	if (res == NULL) {
		fprintf(stderr, "error mysql_query(%s)\n", sql);
		exit(1);
	}
	row = mysql_fetch_row(res);
	if (row && row[0] && strlen(row[0])) {
		FILE *fp = fopen(snapshot_file, "wb");
		fwrite(row[0], strlen(row[0]), 1, fp);
		fclose(fp);
	} else {
		unlink(snapshot_file);
	}

	mysql_close(mysql);
}

// եץSNAPSHOTǡΥȥ
void store_log(struct st_db_desc *db_desc, char *process_id, char *error_dump) {
	MYSQL	*mysql, *con;
	MYSQL_RES	*res;
	MYSQL_ROW	row;
	int	err;
	char	*snapshot;
	char	compress1[16];
	char	compress2[16];

	mysql = mysql_init(NULL);
	if(NULL==mysql){
		fprintf(stderr, "error mysql_init\n");
		exit(1);
	}

	mysql_options(mysql, MYSQL_OPT_COMPRESS, 0);
	con = mysql_real_connect(mysql, db_desc->host,
		db_desc->user, db_desc->pass,
		db_desc->name, atoi(db_desc->port),
		NULL, 0);
	if (NULL==con){
		fprintf(stderr, "error mysql_real_connect\n");
		exit(1);
	}

	// comperssޥɤΥݡ̵ͭå
	if (strcmp(mysql_get_server_info(con), "4.1") >= 0) {
		strcpy(compress1, "compress(");
		strcpy(compress2, ")");
	} else {
		strcpy(compress1, "");
		strcpy(compress2, "");
	}

	// եץλξ硢SNAPSHOTõ
	sprintf(sql, "select END_TIME, ERROR_DUMP from tbl_BOSE_PROC_STAT where PROCESS_ID='%s'", process_id);
	err = mysql_query(mysql, sql);
	if (err != 0) {
		fprintf(stderr, "error mysql_query(%s)\n", sql);
		exit(1);
	}
	res = mysql_store_result(mysql);
	row = mysql_fetch_row(res);

	if (snapshot_mode) {
		snapshot = strdup("");
		if (row[0][0]>'0') {
			unlink(snapshot_file);
		}
	} else {
		if (row[0][0]>'0') {
			snapshot = strdup("");
		} else {
			snapshot = spl_malloc_file(snapshot_file, 0);
		}
		unlink(snapshot_file);
	}

	if (row[1] && strlen(row[1])) {
		error_dump = strdup(row[1]);
	}

	// SNAPSHOTӥγǼեץη³(INWORK_FG)
	char *sql_tmp = malloc((strlen(snapshot) + strlen(error_dump) + 1024) * 2);
	sprintf(sql_tmp, "update tbl_BOSE_PROC_STAT set INWORK_FG='', ERROR_DUMP='%s', SNAPSHOT=%s'%s'%s where PROCESS_ID='%s'", my_escape_string(mysql, error_dump), compress1, my_escape_string(mysql, snapshot), compress2, process_id);
	err = mysql_query(mysql, sql_tmp);
	if (err != 0) {
		fprintf(stderr, "error mysql_query(%s)\n", sql_tmp);
		exit(1);
	}

	// 顼¸ߤ硢QUEǡͭ
	if (strlen(error_dump)) {
		sprintf(sql, "update tbl_BOSE_QUE set INWORK_FG='', ACTIVATE_FG='' where PROCESS_ID='%s'", process_id);

	// ϡեץη³
	} else {
		sprintf(sql, "update tbl_BOSE_QUE set INWORK_FG='' where PROCESS_ID='%s'", process_id);
	}
	err = mysql_query(mysql, sql);
	if (err != 0) {
		fprintf(stderr, "error mysql_query(%s)\n", sql);
		exit(1);
	}

	mysql_close(mysql);
}



int main_proc(char *db_desc_file, char *script_name, char *process_id, FILE *opt_dump_file, FILE *opt_restore_file)
{
	int rc = 0;

	struct spl_vm *vm;
	struct spl_task *task;

	vm = spl_vm_create();
	my_asprintf(&vm->path, ".:./spl_modules:%s", spl_system_modules_dir());
	vm->codecache_dir = 0;
	vm->runloop = spl_simple_runloop;

	// SNAPSHOTѤ³
	if ( opt_restore_file ) {
		if (!spl_restore_ascii(vm, opt_restore_file)) {
			return 1;
		}
		task = vm->task_list;
		if ( opt_dump_file == opt_restore_file ) {
			rewind(opt_dump_file);
			ftruncate(fileno(opt_dump_file), 0);
		} else fclose(opt_restore_file);

	// ưξϡѥ뤵줿ХȥɤѤƼ¹
	} else {
		struct spl_code *code = spl_code_get(0);

		code->code_type = SPL_CODE_MAPPED;
		code->code = spl_mmap_file(script_name, &code->size);

		if (!code->code) {
			spl_report(SPL_REPORT_RUNTIME, vm, "bytecode file not found(%s)", script_name);
			return 1;
		}

		if (*code->code == '#') {
			spl_state_counter_free++;
			code->code_type = SPL_CODE_STATIC;
			while (*code->code != SPL_OP_SIG && code->size > 0) {
				code->code++;
				code->size--;
			}
		}

		task = spl_task_create(vm, "main");
		spl_task_setcode(task, code);
	}

	task->debug = 0;

	// SPLץȤϤ
	int spl_argc = 0;
	struct spl_node *spl_argv = spl_get(0);
	spl_create(task, vm->root, "argv", spl_argv, SPL_CREATE_LOCAL);

	spl_create(task, spl_argv, 0, SPL_NEW_STRING_DUP(db_desc_file), SPL_CREATE_LOCAL); spl_argc++;
	spl_create(task, spl_argv, 0, SPL_NEW_STRING_DUP(process_id), SPL_CREATE_LOCAL); spl_argc++;

	spl_create(task, vm->root, "argc", SPL_NEW_INT(spl_argc), SPL_CREATE_LOCAL);

	// builtinؿϿ
	spl_builtin_register_all(vm);
	spl_clib_reg(vm, "_flow_suspend", handler_flow_suspend, 0);
	spl_clib_reg(vm, "_flow_get_sigterm_info", handler_flow_get_sigterm_info, 0);
	spl_clib_reg(vm, "_flow_get_ppid", handler_flow_get_ppid, 0);
	spl_clib_reg(vm, "_flow_task_create", handler_flow_task_create, 0);
	spl_clib_reg(vm, "_flow_isexist_subprocess", handler_flow_isexist_subprocess, 0);
	spl_clib_reg(vm, "_flow_count_subprocess", handler_flow_count_subprocess, 0);

        spl_clib_reg(vm, "encode_sql", handler_encode_sql, 0);
        spl_clib_reg(vm, "sql_connect", handler_sql_connect_modify, 0);
        spl_clib_reg(vm, "sql", handler_sql, 0);

        spl_hnode_reg(vm, "sql", handler_sqlnode, 0);

	spl_clib_reg(vm, "encode_mysql", handler_encode_mysql, 0);
	sql_register_backend(vm, "mysql", sql_mysql_open_callback);

	// SPLץDispatch롼
	int skip_main = 0;
	while ( 1 ) {
		spl_gc_maybe(vm);

		task = spl_schedule(task);

		// "main"֥ץμ¹ͥٲ
		if (strcmp(task->id, "main")==0) {
			skip_main++;
			if (skip_main > 100) {
				skip_main = 0;
			} else {
				task = spl_schedule(task);
			}
		}
		if ( !task ) break;

		if ( !task->code ||
		     (rc = (spl_exec)(task)) ||
		     got_sigint) break;
	}

	// λSNAPSHOTե񤭽Ф
	if (opt_dump_file) {
		spl_dump_ascii(vm, opt_dump_file);
		fclose(opt_dump_file);
	}

	return rc;
}

int main(int argc, char **argv)
{
	FILE *opt_dump_file = 0;
	FILE *opt_restore_file = 0;
	FILE *fp;
	struct st_db_desc db_desc;

	if (argc < 5) {
		fprintf(stderr, "usage: bose_proc database-def-file script-file process-id dump-file\n");
		exit(-1);
	}

	// 顼ݡȽ
	bose_error_dump = strdup("");
	spl_report = bose_spl_report_func;

	// DBǡ
	fp = fopen(argv[1], "r");
	fgets(db_desc.host, sizeof db_desc.host, fp);
	fgets(db_desc.port, sizeof db_desc.port, fp);
	fgets(db_desc.name, sizeof db_desc.name, fp);
	fgets(db_desc.user, sizeof db_desc.user, fp);
	fgets(db_desc.pass, sizeof db_desc.pass, fp);
	fclose(fp);
	db_desc_trim(db_desc.host);
	db_desc_trim(db_desc.port);
	db_desc_trim(db_desc.name);
	db_desc_trim(db_desc.user);
	db_desc_trim(db_desc.pass);

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigterm_handler);

	// SNAPSHOTǼ("-":DBʳ:File)
	snapshot_mode = 0;
	if (strcmp(argv[4], "-")==0) {
		strcpy(snapshot_file, "/tmp/bose.tmp.XXXXXX");
		mkstemp(snapshot_file);
		// SNAPSHOTDBꥹȥ
		restore_log(&db_desc, argv[3]);
	} else {
		snapshot_mode = 1;
		strcpy(snapshot_file, argv[4]);
	}

	// SNAPSHOTեopen
	opt_dump_file = opt_restore_file = fopen(snapshot_file, "r+");
	if (!opt_restore_file) {
		opt_dump_file = fopen(snapshot_file, "w+");
	}

	// SPLץȼ¹
	main_proc(argv[1], argv[2], argv[3], opt_dump_file, opt_restore_file);

	// SNAPSHOTΥȥ
	store_log(&db_desc, argv[3], bose_error_dump);
}
